﻿
/// <reference path="orgAnice.searchControl.d.ts"/>
/// <reference path="orgAnice.listControl.d.ts"/>

/// <summary>The main command manager; only one instance is supported (and required).</summary>
interface IDialogCommandManager {

	// --- Events ---------------------------------------------------------
	DialogAdded: orgAniceModule.orgAniceEvent<IDialogCommandManagerItem>;
	QueryFailed: orgAniceModule.orgAniceEvent<string>;
	QuerySent: orgAniceModule.orgAniceEvent<string>;
	QuerySucceeded: orgAniceModule.orgAniceEvent<string>;
	PropertyChanged: orgAniceModule.orgAniceEvent<string>;
	SelectionChanged: orgAniceModule.orgAniceEvent<IDialogCommandManagerItem>;

	/// <summary>Raised, when the component is ready for service.</summary>
	Ready: orgAniceModule.orgAniceEvent<void>;

	// --- Properties -----------------------------------------------------
	IsSessionStorageSupported: boolean;
	SearchControl: ISearchControl;
	WarnUnsavedDialog: () => boolean;

	// --- Methods --------------------------------------------------------
	//Add(name: string, endpoint: string): IDialogCommandManagerItem;
	AddDialogItem(dialog: IDialogCommandManagerItem);

	/// <summary><c>true</c>, if the current dialog is modified.</summary>
	IsCurrentModified: boolean;
	/// <summary><c>true</c>, if any dialog is modified.</summary>
    IsModified: boolean;
    /// <summary>If <c>true</c>, the Query() command on each dialog added to the manager will be called once
    ///          with the key set to an empty Guid.</summary>
    IsQueryOnAddDialog: boolean;

	/// <summary>This command calls the ClearCommand of the current dialog.</summary>
	Clear(name?: string);

	CurrentDialog: IDialogCommandManagerItem;

	GetSessionValue(name: string): string;

	HandleAjaxError(xhr: JQueryXHR): void;
	HandleAjaxErrorImpl(reason: any): void;

	IgnoreModified(): void;

	Query(id: string);
    QueryDialog(name: string, id: string);
    RequeryAll(exceptDialog: string): void;
	Restore(): void;
	Save(validate?: boolean): void;

	/// <summary>If modified, displays the confirmation dialog and saves the current view model. Executes validation, if requested.</summary>
	/// <param name="validate">(Optional) true to validate. </param>
	/// <returns><c>true</c> if it succeeds, <c>false</c> if it fails.</returns>
	SaveIfModified(validate?: boolean): boolean;

	Select(name: any): IDialogCommandManagerItem;
	SelectByList(list: IListControl): boolean;
	SessionId: string;
	SetModified(): void;

	SetSessionValue(name: string, value: string): void;

	/// <summary>Validates the current dialog.</summary>
	/// <returns>true if validation succeeds, otherwise false.</returns>
	Validate(): boolean;

	// --- Commands ------------------------------------------------------- 
}

/// <summary>The command manager for one particular dialog.</summary>
interface IDialogCommandManagerItem {

	// --- Events ---------------------------------------------------------
	QueryFailed: orgAniceModule.orgAniceEvent<string>;
	QuerySent: orgAniceModule.orgAniceEvent<string>;
	QuerySucceeded: orgAniceModule.orgAniceEvent<string>;

	OnClearCalled: IEvent;
	OnDeleted: IEvent;
	OnQuery: IEvent;
	OnSaved: IEvent;

	PropertyChanged: orgAniceModule.orgAniceEvent<string>;

	RelationsChanged: orgAniceModule.orgAniceEvent<IRelationsChangedEventArgs>;

	ValidationStateChanged: orgAniceModule.orgAniceEvent<boolean>;

	// --- Properties -----------------------------------------------------

	CanCreate: boolean;

	/// <summary>Gets or sets a value indicating, whether this instance can delete its <see cref="ViewModel"/>. </summary>
	/// <returns><c>true</c>, if values can be deleted; otherwise, <c>false</c>. </returns>
	CanDelete: boolean;

	/// <summary>Gets or sets a value indicating, whether this instance can save its <see cref="ViewModel"/>. </summary>
	/// <returns><c>true</c>, if values can be deleted; otherwise, <c>false</c>. </returns>
	CanSave: boolean;

	/// <summary>Gets the name of the controller which is responsible for servicing the requests.</summary>
	/// <remarks>This property should be initialized when the item is added, and be read-only during operation.</remarks>
	Controller: string;

	/// <summary>Returns the OID of the current data object, which will be set by the Query method.< / summary >
	/// <returns>The OID of the current data object.</returns>
	CurrentOid: string;

	IsFillMode: boolean;
    IsModified: boolean;
	IsNew: boolean;
	IsReadonly: boolean;

	/// <summary><c>true</c> if this dialog is selected; otherwise, <c>false</c>.</summary>
	IsSelected: boolean;

	/// <summary>The list control associated with the frame/dialog.</summary>
	/// <remarks>The list which is associated with the current frame. A <see cref="IListControl"/> manages its own
	///			 commands.</remarks>
	List: IListControl;

    Name: string;
    Parent: IDialogCommandManager;

    /// <summary>The data endpoint; usually, this is the name of the service controller at the server.</summary>
    ServiceEndpoint: string;

	/// <summary>The validator which define validation rules on the properties of the ViewModel.</summary>
	TypeValidator: IOrgDataTypeValidator;

	Validators: IDialogValidator[];

	ValidationHasError: boolean;

	ViewModel: {};

	// --- Methods --------------------------------------------------------
	ApplyValidations(): void;

	Clear(): void;
	Delete(): void;
	FinishDeleted(): void;

	GetSessionValue(name: string): string;

	IndexChanged(args: IIndexChangedEventArgs): void;
    Query(oid: string): void;
    QueryIndexes(): void;

	Restore(): void;

	ResumePropertyChanged(): void;

	Save(validate?: boolean): void;

	SaveIfModified(validate?: boolean): boolean;

	Search(options): void;

	SetList(list: IListControl): void;

	SetSessionValue(name: string, value: string): void;

	SuspendPropertyChanged(): void;

	/// <summary>Returns this instance' properties as a plain object. </summary>
	/// <returns>This instance' properties converted to a plain model. </returns>
	ToViewModel(): {};

	/// <summary>Validates the dialog.</summary>
	/// <returns>true if validation succeeds, otherwise false.</returns>
	Validate(): boolean;

	// --- Commands -------------------------------------------------------
	/// <summary>This command clears all fields in the dialog.</summary>
	ClearCommand: (options: any) => void;

	/// <summary>The command which deletes the entry given by its <c>id</c>.</summary>
	DeleteCommand: (options: {}) => void;

	FillViewModel(data: {}): void;

	FillViewModelCommand: (data: {}) => void;

	Initialize: () => void;

	InitializeCommand: () => void;

	/// <summary>The query command, used to query a single entry by ID.</summary>
	QueryCommand: (id: string) => void;

	RestoreCommand: () => void;

	/// <summary>The save commands of the dialogs.</summary>
	SaveCommand: (options: {}) => void;
}

/// <summary>Exposes the state and text of a single validation rule.</summary>
interface IDialogValidator {
	DialogManagerItem: IDialogCommandManagerItem;
	ErrorText: string;
	HasError: boolean;
}
