﻿
///<reference path="../typings/jquery/jQuery.d.ts"/>
///<reference path="orgAnice.d.ts"/>

interface INavigationViewModel {
	HomeOid: KnockoutObservable<string>;	// points to address
	BackOid: KnockoutObservable<string>;	// points to where we came from
	NextOid: KnockoutObservable<string>;	// points to the next logical page
	CurrentOid: KnockoutObservable<string>;	// points to the current id in process

	HomeLink: string;
	BackLink: string;
	BackText: string;
	NextLink: string;
	NextText: string;

	BackUrl: string;
	HomeUrl: string;
	NextUrl: string;

	HasCurrentOid: boolean;
	HasBackLink: boolean;
	HasBackUrl: boolean;
	HasHomeLink: boolean;
	HasHomeUrl: boolean;
	HasNextLink: boolean;
	HasNextUrl: boolean;
	HasBackOid: boolean;
	HasNextOid: boolean;

	UserPrefix: string;
}

interface INavigationControl {
	// --- Events ---------------------------------------------------------
	PropertyChanged: orgAniceModule.IorgAniceEvent<string>;
	Ready: orgAniceModule.IorgAniceEvent<string>;

	// --- Properties -----------------------------------------------------
	IsReady: boolean;
	Orientation: orgAniceModule.Orientation;

	// --- Methods --------------------------------------------------------
	FindByName(name: string): INavigationItem;
	Initialize(options: INavigationOptions);
	ResumePropertyChanged();
	SuspendPropertyChanged();
}

interface INavigationOptions {
	Items?: INavigationItemOptions[];
	Orientation?: orgAniceModule.Orientation;
}

interface INavigationItemOptions {
	DisplayText?: string;
	EventType: string;
	Img?: string;
	ImgActive?: string;
	ImgDisabled?: string;
	IsChecked: boolean;
	IsEnabled?: boolean;
	IsToggle: boolean;
	IsVisible?: boolean;
	Name: string;
}

interface INavigationItem {
	// --- Events ---------------------------------------------------------
	/// <summary>Triggered, if item is a toggle button, and the checked state has changed.</summary>
	CheckedChanged: orgAniceModule.IorgAniceEvent<boolean>;

	/// <summary>Triggered, if the item is clicked.</summary>
	Click: orgAniceModule.orgAniceEvent<string>;

	/// <summary>Triggered, if any property on the item has changed.</summary>
	PropertyChanged: orgAniceModule.IorgAniceEvent<string>;

	// --- Properties -----------------------------------------------------

	/// <summary>The text which is displayed for the item; usually translated to the user's language.</summary>
	DisplayText: string;

	/// <summary>Full pathname of the image file; this may change according to the item's state.</summary>
	/// <remarks>According to the item's state (IsEnabled, IsToggleButton), the path may change.</remarks>
	ImagePath: string;

	/// <summary><c>true</c> if this item acts like a toggle button, and the button is checked; otherwise, <c>false</c>.</summary>
	IsChecked: boolean;

	/// <summary><c>true</c> if this item is enabled (clickable/touchable); otherwise, <c>false</c>.</summary>
	IsEnabled: boolean;

	/// <summary><c>true</c> if this item acts like a toggle button; otherwise, <c>false</c>.</summary>
	IsToggleButton: boolean;

	/// <summary><c>true</c> if this item is visible; otherwise, <c>false</c>.</summary>
	IsVisible: boolean;

	Name: string;

	/// <summary>A function that will be called on activation of the navigation item; determines, whether the action may
	///			 be applied (return value: <c>true</c>) or not (return value: <c>false</c>). </summary>
	/// <returns><c>true</c>, if hte action may be applied; otherwise <c>false</c>. </returns>
	ValidateAction: () => boolean;

	// --- Methods --------------------------------------------------------
	SetEvent(options: {
		AddRemove: boolean; // true: add; false: remove
		EventName: string;
		EventFunc: (args?: string) => {};
	});

	// ===========================================================================
	// ===========================================================================
	// TODO: Find "something" to delay setting properties to after initialization!
	// ===========================================================================
	// ===========================================================================
	SetProperty(options: {
		IsEnabled?: boolean;
		IsVisible?: boolean;
	});
}

class NavigationControl implements INavigationControl {
	// --- Events ---------------------------------------------------------
	public PropertyChanged: orgAniceModule.IorgAniceEvent<string> = new orgAniceModule.orgAniceEvent<string>();
	public Ready: orgAniceModule.IorgAniceEvent<string> = new orgAniceModule.orgAniceEvent<string>();

	// --- Fields ---------------------------------------------------------
	private isReadyFired: boolean = false;
	private items: INavigationItem[];
	private orientation: orgAniceModule.Orientation = orgAniceModule.Orientation.Horizontal;
	private propertyChangedSuspended: boolean = false;
	private topLevel: any;
	private viewModel: {
		Items: INavigationItem[];
		Orientation: orgAniceModule.Orientation;
	} = {
		Items: [],
		Orientation: orgAniceModule.Orientation.Horizontal
	}

	// --- Properties -----------------------------------------------------
	public get IsReady(): boolean { return this.isReadyFired; }

	public get Orientation(): orgAniceModule.Orientation { return this.orientation; }

	public set Orientation(value: orgAniceModule.Orientation) {
		this.orientation = value;
		this.onPropertyChanged("Orientation");
	}

	// --- Methods --------------------------------------------------------
	public Initialize(options: INavigationOptions) {
		this.create(options);
	}

	public FindByName(name: string): INavigationItem {
		for (var idx = 0; idx < this.viewModel.Items.length; ++idx) {
			var item = this.viewModel.Items[idx];
			if (item != undefined && item.Name == name)
				return item;
		}
		return undefined;
	}

	public ResumePropertyChanged() {
		this.propertyChangedSuspended = false;
	}

	public SuspendPropertyChanged() {
		this.propertyChangedSuspended = true;
	}

	// --- Constructor ----------------------------------------------------
	constructor(placeholder: any, options: INavigationOptions) {
		if (placeholder == undefined)
			throw "The toplevel element (placeholder in HTML) must be assigned!";

		if (placeholder.length == 0)
			throw "The toplevel element is not a valid jQuery selector!";

		this.topLevel = $(placeholder);

		this.create(options);
	}

	// --- Implementation -------------------------------------------------
	private create(options: INavigationOptions) {

		this.viewModel.Orientation = options.Orientation;

		// protect against early calls, like attaching events and so on
		if (options.Items == undefined || options.Items.length == 0) return;

		var container = this.topLevel;
		container.empty();
		this.topLevel.addClass(this.viewModel.Orientation == orgAniceModule.Orientation.Horizontal ? "horizontal-pane" : "vertical-pane");

		var outerBorder = $(document.createElement("div"));
		outerBorder.addClass("outer-border");

		var innerBorder = $(document.createElement("div"));
		innerBorder.addClass("inner-border");

		var frame = $(document.createElement("div"));
		frame.addClass("nav-pane");

		innerBorder.append(frame);
		outerBorder.append(innerBorder);
		container.append(outerBorder);

		$.each(options.Items, (index, value) => {
			var navItem = new NavigationItem(value.Name, value);
			this.viewModel.Items.push(navItem);

			frame.append(navItem.DomElement);
		});

		this.onReady();
	}

	private onPropertyChanged(property: string) {
		if (this.propertyChangedSuspended) return;
		this.PropertyChanged.trigger(this, property);
	}

	private onReady() {
		this.isReadyFired = true;

		this.Ready.FireOnAdd(this.isReadyFired);

		this.Ready.trigger("");
	}
}

class NavigationItem implements INavigationItem {
	// --- Events ---------------------------------------------------------
	public CheckedChanged: orgAniceModule.IorgAniceEvent<boolean> = new orgAniceModule.orgAniceEvent<boolean>();

	public Click: orgAniceModule.orgAniceEvent<string> = new orgAniceModule.orgAniceEvent<string>();

	public PropertyChanged: orgAniceModule.IorgAniceEvent<string> = new orgAniceModule.orgAniceEvent<string>();

	// --- Fields ---------------------------------------------------------
	private domElement: any;
	private eventType: any;
	private imagePath: string;
	private imagePathDisabled: string;
	private imagePathActive: string;
	private isChecked: boolean;
	private isEnabled: boolean;
	private isToggle: boolean;
	private name: string;
	private supendEvents: boolean = true;	// suspend until constructor is called
	private validation: () => boolean;
	private viewModel: {
		DisplayText: KnockoutObservable<string>;
		ImagePath: KnockoutObservable<string>;
		IsVisible: KnockoutObservable<boolean>;
		Name: KnockoutObservable<string>;
		OnClick: Function;
	} = {
		DisplayText: ko.observable<string>(),
		ImagePath: ko.observable<string>(),
		IsVisible: ko.observable<boolean>(),
		Name: ko.observable<string>(this.name),
		OnClick: () => {
			this.onClick();
		}
	};

	// --- Properties -----------------------------------------------------
	public get DisplayText(): string { return this.viewModel.DisplayText(); }

	public set DisplayText(value: string) {
		this.viewModel.DisplayText(value);
		this.onPropertyChanged("DisplayText");
	}

	public get DomElement(): any { return this.domElement; }

	public get ImagePath(): string { return this.imagePath; }

	public set ImagePath(value: string) {
		this.imagePath = value;
		this.onPropertyChanged("ImagePath");

		if (this.IsEnabled && this.viewModel.IsVisible())
			this.viewModel.ImagePath(this.imagePath);
	}

	public get ImagePathActive(): string { return this.imagePathActive; }

	public set ImagePathActive(value: string) {
		this.imagePathActive = value;
		this.onPropertyChanged("ImagePathActive");
	}

	public get ImagePathDisabled() { return this.imagePathDisabled; }

	public set ImagePathDisabled(value: string) {
		this.imagePathDisabled = value;
		this.onPropertyChanged("ImagePathDisabled");

		if (!this.isEnabled)
			this.viewModel.ImagePath(this.imagePathDisabled);
	}

	public get IsChecked(): boolean { return this.isChecked; }

	public set IsChecked(value: boolean) {
		if (this.isChecked == value) return;

		this.isChecked = value;

		this.toggleVisualState(this.isChecked, "active", this.imagePathActive, false);

		this.onPropertyChanged("IsChecked");
		this.onCheckChanged(this.isChecked);
	}

	public get IsEnabled(): boolean { return this.isEnabled; }

	public set IsEnabled(value: boolean) {
		this.isEnabled = value;

		this.toggleVisualState(this.isEnabled, "disabled", this.imagePathDisabled, true);

		this.onPropertyChanged("IsEnabled");
	}

	public get IsToggleButton():boolean { return this.isToggle; }

	public get IsVisible(): boolean { return this.viewModel.IsVisible(); }

	public set IsVisible(value: boolean) {
		this.viewModel.IsVisible(value);
		this.onPropertyChanged("IsVisible");
	}

	public get Name(): string { return this.name; }

	public get ValidateAction(): () => boolean { return this.validation; }

	public set ValidateAction(value: () => boolean) { this.validation = value; }

	// --- Methods --------------------------------------------------------
	public SetEvent(options: {
		AddRemove: boolean; // true: add; false: remove
		EventName: string;
		EventFunc: any;
	}) {
		switch (options.EventName.toLowerCase()) {
			case "click":
				if (typeof options.AddRemove == "boolean" && options.AddRemove)
					this.Click.add(options.EventFunc);
				else
					this.Click.remove(options.EventFunc);
				break;

			case "checkchanged":
				if (typeof options.AddRemove == "boolean" && options.AddRemove)
					this.CheckedChanged.add(options.EventFunc);
				else
					this.CheckedChanged.remove(options.EventFunc);
				break;
		}
	}

	public SetProperty(options: {}) {
		var instance = this;
		$.each(options, (index, value) => {
			if (instance.viewModel.hasOwnProperty(index)) {
				if (typeof instance.viewModel[index] == "function")
					instance.viewModel[index].call(instance, value);
				else
					instance.viewModel[index] = value;
			} else if (instance.hasOwnProperty(index)) {
				instance[index] = value;
				instance.onPropertyChanged(index);
			} else {
				switch (index.toLowerCase()) {
					case "displaytext":
						instance.DisplayText = value;
						break;
					case "ischecked":
						instance.IsChecked = value;
						break;
					case "isenabled":
						instance.IsEnabled = value;
						break;
					case "isvisible":
						instance.IsVisible = value;
						break;
					case "name":
						instance.viewModel.Name(value);
						break;
				}
			}
		});
	}

	// --- Constructor ----------------------------------------------------
	constructor(name: string, options: INavigationItemOptions) {
		this.name = name;
		this.eventType = options.EventType;
		this.imagePath = options.Img;
		this.imagePathActive = options.ImgActive;
		this.imagePathDisabled = options.ImgDisabled;
		this.isChecked = options.IsChecked || false;
		this.isEnabled = options.IsEnabled || true;
		this.isToggle = options.IsToggle || false;
		this.IsVisible = options.IsVisible || true;

		this.viewModel.Name(this.name);
		this.viewModel.DisplayText(options.DisplayText == undefined ? this.name : options.DisplayText);

		this.domElement = this.create(options);

		ko.applyBindings(this.viewModel, this.domElement[0]);

		this.onPropertyChangedInternal("IsChecked");
		this.onPropertyChangedInternal("IsEnabled");

		this.supendEvents = false;
	}

	// --- Implementation -------------------------------------------------
	private create(options: INavigationItemOptions): any {
		var container = $(document.createElement("div"));
		container.addClass("nav-item");
		container.attr("data-bind", "visible: IsVisible, attr: { id: Name }, click: OnClick");

		var imgdiv = $(document.createElement("div"));
		imgdiv.addClass("imgdiv");

		var txt = $(document.createElement("div"));
		txt.addClass("txtdiv");
		txt.attr("data-bind", "text: DisplayText");

		container.append(imgdiv);
		container.append(txt);

		return container;
	}

	private onCheckChanged(value: boolean) {
		if (this.supendEvents) return;

		this.CheckedChanged.trigger(value);
	}

	private onClick() {
		if (this.supendEvents || !this.isEnabled) return;

		// if a validation is assigned, but fails -> do nothing
		if (this.validation != undefined && !this.validation()) return;

		if (this.IsToggleButton) {
			this.IsChecked = !this.isChecked;
		}

		if (this.eventType != undefined)
			eval("orgAnice.trigger(this.eventType);");

		//TODO: "eval()" is not the way to go! Changed this to a reference to the "orgAnice" namespace
	}

	private onPropertyChanged(property: string) {
		if (this.supendEvents) return;

		this.onPropertyChangedInternal(property);

		this.PropertyChanged.trigger(this, property);
	}

	private onPropertyChangedInternal(property: string) {
		var imagePath;
		switch (property) {
			case "ImagePathActive":
				break;

			case "IsEnabled":
				imagePath = this.isEnabled
					? (this.isToggle && this.isChecked && this.imagePathActive != undefined ? this.imagePathActive : this.imagePath)
					: (this.imagePathDisabled != undefined ? this.imagePathDisabled : this.imagePath);
				this.viewModel.ImagePath(imagePath);
				break;

			case "IsChecked":
				imagePath = this.isEnabled
					? (this.isChecked && this.imagePathActive != undefined ? this.imagePathActive : this.imagePath)
					: (this.imagePathDisabled != undefined ? this.imagePathDisabled : this.imagePath);
				this.viewModel.ImagePath(imagePath);
				if (this.isChecked)
					this.domElement.addClass("checked");
				else
					this.domElement.removeClass("checked");
				break;
		}
	}

	private toggleVisualState(toggle: boolean, selector: string, alternateImage: string, reverseProperty: boolean) {

		this.viewModel.ImagePath(toggle && alternateImage != undefined ? alternateImage : this.imagePath);

		if ((toggle && !reverseProperty) || (!toggle && reverseProperty)) {
			this.domElement.addClass(selector);
			this.domElement.find(".imgdiv").addClass(selector);
			this.domElement.find(".txtdiv").addClass(selector);
		} else {
			this.domElement.removeClass(selector);
			this.domElement.find(".imgdiv").removeClass(selector);
			this.domElement.find(".txtdiv").removeClass(selector);
		}
	}
}

class NavigationViewModel implements INavigationViewModel {
	// --- Constant Values ------------------------------------------------
	public static EmptyGuid: string = "00000000-0000-0000-0000-000000000000";
	public static EmptyRequestGuid: string = "00000000-0000-0000-0000-000000000001";

	public static HomeCookieOid: string = "addressOid";
	public static BackCookieOid: string = "sourceOid";
	public static NextCookieOid: string = "targetOid";
	public static CurrentCookieOid: string = "currentOid";

	// --- Fields ---------------------------------------------------------
	private homeOid: KnockoutObservable<string> = ko.observable("");
	private backOid: KnockoutObservable<string> = ko.observable("");
	private nextOid: KnockoutObservable<string> = ko.observable("");
	private currentOid: KnockoutObservable<string> = ko.observable("");

	private homeLink: string;
	private homeText: string;
	private backLink: string;
	private backText: string;
	private isBackVisible: boolean;
	private isHomeVisible: boolean;
	private isNextVisible: boolean;
	private nextLink: string;
	private nextText: string;
	private nextCookieName: string;

	// --- Properties -----------------------------------------------------

	/// <summary>The Oid of the current entity in the active display.</summary>
	/// <returns>A KnockoutObservable&lt;string&gt;</returns>
	public get CurrentOid(): KnockoutObservable<string> { return this.currentOid; }

	/// <summary>The Oid of the entity, which shall be loaded by the page, where the <i>Home</i> button points to.</summary>
	/// <returns>A KnockoutObservable&lt;string&gt;</returns>
	/// <remarks>Currently ,this is usually the <i>Address</i> page.</remarks>
	public get HomeOid(): KnockoutObservable<string> { return this.homeOid; }

	/// <summary>The Oid of the entity, which shall be loaded by the page, where the <i>Back</i> button points to.</summary>
	/// <returns>A KnockoutObservable&lt;string&gt;</returns>
	public get BackOid(): KnockoutObservable<string> { return this.backOid; }

	/// <summary>The Oid of the entity, which shall be loaded by the page, where the <i>Next</i> button points to.</summary>
	/// <returns>A KnockoutObservable&lt;string&gt;</returns>
	public get NextOid(): KnockoutObservable<string> { return this.nextOid; }

	public get BackLink(): string { return this.backLink; }
	public set BackLink(value: string) { this.backLink = value; }

	public get BackText(): string { return this.backText; }

	public get HomeLink(): string { return this.homeLink; }
	public set HomeLink(value: string) { this.homeLink = value; }

	public get HomeText(): string { return this.homeText; }

	public get IsBackVisible(): boolean { return this.isBackVisible; }
	public get IsHomeVisible(): boolean { return this.isHomeVisible; }
	public get IsNextVisible(): boolean { return this.isNextVisible; }

	public get NextLink(): string { return this.nextLink; }
	public set NextLink(value: string) { this.nextLink = value; }

	public get NextText(): string { return this.nextText; }

	public get TargetCookieName(): string { return this.nextCookieName; }

	public get BackUrl(): string {
		return (this.HasBackLink && this.HasBackOid)
			? this.BackLink // + "/" + this.BackOid()
			: "";
	}

	public get HomeUrl(): string {
		return (this.HasHomeLink && this.HasHomeOid)
			? this.HomeLink // + "/" + this.HomeOid()
			: "";
	}

	public get NextUrl(): string {
		return (this.HasNextLink && this.HasNextOid)
			? this.NextLink // + "/" + this.NextOid()
			: "";
	}

	public get HasBackLink(): boolean { return this.backLink != undefined && this.backLink != ""; }
	public get HasBackUrl(): boolean {
		return this.HasBackLink && this.HasBackOid && this.BackOid() != NavigationViewModel.EmptyRequestGuid;
	}
	public get HasBackOid(): boolean {
		return this.backOid() != undefined && this.backOid().length > 0 && this.backOid() != NavigationViewModel.EmptyRequestGuid;
	}

	public get HasCurrentOid(): boolean {
		return orgAniceModule.IsValidGuid(this.currentOid()) && !orgAniceModule.IsEmptyGuid(this.currentOid());
	}

	public get HasHomeLink(): boolean { return this.homeLink != undefined && this.homeLink != ""; }
	public get HasHomeOid(): boolean { return this.homeOid() != undefined && this.homeOid().length > 0; }
	public get HasHomeUrl(): boolean {
		return this.HasHomeOid && this.HasHomeLink && this.homeOid() != NavigationViewModel.EmptyRequestGuid;
	}

	public get HasNextLink(): boolean { return this.nextLink != undefined && this.nextLink != ""; }
	public get HasNextUrl(): boolean { return this.HasNextLink && this.HasNextOid; }
	public get HasNextOid(): boolean { 
		return !this.IsEmptyGuid(this.nextOid());
		//return this.nextOid() != undefined && this.nextOid().length > 0;
	}

	public get UserPrefix(): string {
		var cookie = $.cookie("orgAniceUser");
		return cookie == undefined ? "unknown" : cookie;
	}

	// --- Methods --------------------------------------------------------
	public IsEmptyGuid(value: any): boolean {
		return value == undefined || value === "" || value === NavigationViewModel.EmptyGuid;
	}

	// --- Constructor ----------------------------------------------------
	constructor(json?: string) {
		if (json == undefined) return;

		this.backLink = json["BackLink"];
		this.backText = json["BackText"] || "Zurück";
		this.nextLink = json["NextLink"];
		this.nextText = json["NextText"] || "Weiter";
		this.homeLink = json["HomeLink"];
		this.homeText = json["HomeText"];
		this.nextCookieName = json["NextCookieName"] || "NextOid";

		var userCookie = this.UserPrefix;

		var cookie = $.cookie(userCookie + NavigationViewModel.HomeCookieOid);
		this.homeOid(cookie == undefined ? NavigationViewModel.EmptyRequestGuid : cookie);

		cookie = $.cookie(userCookie + NavigationViewModel.CurrentCookieOid);
		this.currentOid(cookie == undefined ? NavigationViewModel.EmptyRequestGuid : cookie);

		cookie = $.cookie(userCookie + NavigationViewModel.BackCookieOid);
		this.backOid(cookie == undefined ? NavigationViewModel.EmptyRequestGuid : cookie);

		cookie = $.cookie(userCookie + NavigationViewModel.NextCookieOid);
		this.nextOid(cookie == undefined ? NavigationViewModel.EmptyRequestGuid : cookie);

		this.homeOid.subscribe((oid: string) => {
			$.cookie(userCookie + NavigationViewModel.HomeCookieOid, oid, { path: '/', expires: 7 });
		});

		this.isBackVisible = this.getBoolFromJson("IsBackVisible", true, json);
		this.isHomeVisible = this.getBoolFromJson("IsHomeVisible", true, json);
		this.isNextVisible = this.getBoolFromJson("IsNextVisible", true, json);
	}

	// --- Implementation -------------------------------------------------
	private getBoolFromJson(name: string, fallback: boolean, json: string): boolean {
		var value = json[name];
		return (value != undefined) ? value : fallback;
	}
}

// ReSharper disable JsFunctionCanBeConvertedToLambda
(function ($) {
	$.fn.orgAniceNavigation = function (arg1: any, arg2: any, arg3: any) {

		if (!this.length) return this;

		var defaults: INavigationOptions = {
			Orientation: orgAniceModule.Orientation.Horizontal
		}

		var count = arguments.length;

		return this.each(function() {
			var instance = this;
			var $instance = $(instance);
			var options: INavigationOptions;
			var control: INavigationControl;

			switch (count) {
			case 0:
				return;

			case 1:
				options = $.extend({}, arg1, defaults);
				control = $instance.data["orgAniceNavigation"];

				if (control == undefined) {
					control = new NavigationControl(this, options);
					$instance.data["orgAniceNavigation"] = control;
				} else {
					control.Initialize(options);
				}
				break;

			default: // arguments.length > 1
				var target: string = arg1;
				var selector: string = arg2;
				var value: any = arg3;
				var item;

				control = $instance.data["orgAniceNavigation"];
				if (control == undefined) {
					control = new NavigationControl(this, defaults);
					$instance.data["orgAniceNavigation"] = control;
				}

				switch (target.toLowerCase()) {
				case "event":
					if (selector.toLowerCase() == "ready") {
						control.Ready.add(value);
						break;
					}

					item = control.FindByName(selector);
					if (item == undefined) return;

					item.SetEvent(value);

					break;

				case "property":
					selector = selector.replace(/\{/, '').replace(/\}/, '');
					var items = selector.split(/[ \,\;]/);
					$.each(items, (index, navItem) => {
						item = control.FindByName(navItem);
						if (item != undefined)
							item.SetProperty(value);
					});
					break;
				}

				break;
			}
		});
	};
})(jQuery || $);
// ReSharper restore JsFunctionCanBeConvertedToLambda
