﻿
///<reference path="../typings/jquery/jquery.d.ts"/>
///<reference path="../typings/knockout-3.2.0.d.ts"/>
///<reference path="orgAnice.popupManager.d.ts"/>

/// <summary>Manager for dialogs.</summary>
/// <seealso cref="T:IPopupManager"/>
class PopupManager implements IPopupManager {
	// --- Events ---------------------------------------------------------
	// --- Fields ---------------------------------------------------------
	private dialogs: { [index: string]: PopupItem };

	// --- Properties -----------------------------------------------------
	public get Dialogs(): { [index: string]: PopupItem } { return this.dialogs; }

	// --- Methods --------------------------------------------------------
	public AddDialog(json: {}): PopupItem {

		if (json == undefined || !json.hasOwnProperty("Name")) return null;

		var name = json["Name"];
		var dialogItem = new PopupItem(name);

		dialogItem.FromJson(json);

		this.dialogs[name] = dialogItem;

		return dialogItem;
	}

	public Show(name: string): PopupItem {
		var item = this.dialogs[name];

		if (item != undefined)
			item.Show();

		return item;
	}

	// --- Constructor ----------------------------------------------------
	// --- Implementation -------------------------------------------------
}

class PopupItem implements IPopupItem {
	// --- Events ---------------------------------------------------------
	// --- Fields ---------------------------------------------------------
	private buttons: DialogButton[] = [];
	private captionText: string;
	private contentName: string;
	private height: number;
	private invoker: string;
	private isAllowScroll: boolean;
	private isAutoOpen: boolean;
	private isModal: boolean;
	private isPopup: boolean;
	private isResizable: boolean;
	private name: string;
	private text: string;
	private width: number;

	// --- Properties -----------------------------------------------------
	public get IsAllowScroll(): boolean { return this.isAllowScroll != undefined && this.isAllowScroll; }

	public set IsAllowScroll(value: boolean) { this.isAllowScroll = value; }

	/// <summary>Gets the buttons of the dialog.</summary>
	/// <value>The buttons of the dialog.</value>
	/// <remarks>The buttons will be rendered from left to right in the order they appear in the list.</remarks>
	public get Buttons(): DialogButton[] { return this.buttons; }

	/// <summary>Gets or sets the caption text.</summary>
	/// <value>The caption text.</value>
	public get CaptionText(): string { return this.captionText; }

	public set CaptionText(text: string) { this.captionText = text; }

	/// <summary>Gets or sets the name of the content; see <b>Remarks</b> for details.</summary>
	/// <value>The name of the content.</value>
	/// <remarks>The <see cref="ContentName"/> is the name of a partial control, which may be rendered inside the dialog.
	/// 		 It will be displayed between the <see cref="Text"/> (if any) and the <see cref="Buttons"/> of the
	/// 		 dialog. If a <b>Content</b> is given, the <see cref="Data"/> will be passed to the partial control.</remarks>
	public get ContentName(): string { return this.contentName; }

	public set ContentName(name: string) { this.contentName = name; }

	public get Height(): number { return this.height; }

	public set Height(height: number) { this.height = height; }

	public get Invoker(): Function {
		return $.fn[this.invoker];
	}

	public get IsAutoOpen(): boolean { return this.isAutoOpen != undefined && this.isAutoOpen; }

	public set IsAutoOpen(value: boolean) { this.isAutoOpen = value; }

	/// <summary>Gets or sets a value indicating whether the dialog is modal.</summary>
	/// <value><c>true</c> if the dialog is modal, <c>false</c> if not.</value>
	public get IsModal(): boolean { return this.isModal; }

	public set IsModal(isModal: boolean) { this.isModal = isModal; }

	/// <summary>Gets or sets a value indicating whether to create a popup dialog.</summary>
	/// <value><c>true</c> if this is a popup dialog, <c>false</c> if not.</value>
	public get IsPopup(): boolean { return this.isPopup; }

	public set IsPopup(isPopup: boolean) { this.isPopup = isPopup; }

	public get IsResizable(): boolean { return this.isResizable; }

	public set IsResizable(value: boolean) { this.isResizable = value; }

	public get Name(): string { return this.name; }

	/// <summary>Gets or sets the text.</summary>
	/// <value>The text.</value>
	public get Text(): string { return this.text; }

	public get Width(): number { return this.width; }

	public set Width(width: number) { this.width = width; }

	// --- Methods --------------------------------------------------------
	public FromJson(json: {}): PopupItem {

		if (json.hasOwnProperty("CaptionText"))
			this.captionText = json["CaptionText"];
		if (json.hasOwnProperty("ContentName"))
			this.contentName = json["ContentName"];
		if (json.hasOwnProperty("IsModal"))
			this.isModal = json["IsModal"];
		if (json.hasOwnProperty("IsPopup"))
			this.isModal = json["IsPopup"];
		if (json.hasOwnProperty("Name"))
			this.name = json["Name"];
		if (json.hasOwnProperty("Text"))
			this.text = json["Text"];
		if (json.hasOwnProperty("Height")) {
			var height = json["Height"];
			this.height = height != undefined && height > 0 ? height : undefined;
		}
		if (json.hasOwnProperty("Width")) {
			var width = json["Width"];
			this.width = width != undefined && width > 0 ? width : undefined;
		}

		if (json.hasOwnProperty("Buttons")) {
			$.each(json["Buttons"], (index, value) => {
				var button = new DialogButton();
				button.FromJson(value);
				this.buttons.push(button);
			});
		}

		return this;
	}

	public Initialize() {

		var root = $("#" + this.name);
		if (root == undefined) {
			//orgAnice.reportError("Dialog " + this.name + " nicht gefunden");
			return;
		}

		var buttons = [];
		if (this.buttons != undefined && this.buttons.length > 0) {

			$.each(this.buttons, (index, button) => {
				var name = "popupButton_" + this.name + "_" + button.Name;
				var btn = {
					text: button.Text,
					click: function() {
						$(this).dialog("close");
						if ($.fn[name] == undefined && button.Command != undefined)
							$.fn[name] = button.Command;
					}
				}
				buttons.push(btn);
			});
		}

		root.dialog(this.createOptions(buttons));

		//root.dialog({
		//	autoOpen: this.IsAutoOpen,
		//	resizable: true, // this.IsResizable != undefined && this.IsResizable,
		//	modal: this.IsModal == undefined || this.isModal
		//});

		this.invoker = "$.fn.openPopup_" + this.name;

		$.fn[this.invoker] = () => { eval("$(\"# + this.name\").dialog(\"open\");"); }
	}

	public Show() {
		eval("$(\"#" + this.name + "\").dialog(\"open\");");
	}

	// --- Constructor ----------------------------------------------------
	constructor(name: string, jsonInit: {} = null) {
		this.name = name;
		if (jsonInit != null)
			this.FromJson(jsonInit);
	}

	// --- Implementation -------------------------------------------------
	private createOptions(buttons: DialogButton[]): {} {
		var options = {};
		options["autoOpen"] = this.IsAutoOpen;
		options["resizable"] = true;

		if (this.height != undefined && this.height > 0)
			options["height"] = this.height;

		if (this.width != undefined && this.width > 0)
			options["width"] = this.width;

		if (this.buttons.length > 0)
			options["buttons"] = buttons;

		if (!this.IsAllowScroll) {
			options["open"] = () => {
				$("#" + this.Name).css("overflow", "hidden");
			}
		}

		return options;
	}


	private toLowerCamelCase(value: string): string {
		if (value == undefined) return value;

		var result = value[0].toLowerCase();
		result += value.substring(1);

		return result;
	}
}

class DialogButton {
	// --- Events ---------------------------------------------------------
	// --- Fields ---------------------------------------------------------
	private command: Function;
	private isCloseDialog: boolean;
	private isDefaultButton: boolean;
	private name: string;
	private text: string;

	// --- Properties -----------------------------------------------------
	public get Command() { return this.command; }

	public set Command(cmd: Function) { this.command = cmd; }

	public get IsCloseDialog(): boolean { return this.isCloseDialog != undefined && this.isCloseDialog; }

	public get IsDefaultButton(): boolean { return this.isDefaultButton != undefined && this.isDefaultButton; }

	public set IsDefaultButton(value: boolean) { this.isDefaultButton = value; }

	public get Name(): string {
		var name = this.name;

		if (name == undefined || name.length == 0)
			this.name = name = this.Text || "__X__";

		return name.replace(" ", "_");
	}

	public set Name(value: string) { this.name = value; }

	public get Text(): string { return this.text; }

	// --- Methods --------------------------------------------------------
	public FromJson(json: {}): DialogButton {

		if (json.hasOwnProperty("Text"))
			this.text = json["Text"];
		if (json.hasOwnProperty("IsCloseDialog"))
			this.isCloseDialog = json["IsCloseDialog"];
		if (json.hasOwnProperty("IsDefaultButton"))
			this.isDefaultButton = json["IsDefaultButton"];

		return this;
	}

	// --- Constructor ----------------------------------------------------
	// --- Implementation -------------------------------------------------
}